(** Infrastructure and tactics for using a depth measure on collapsed syntax

    Author: Anne Mulhern (mulhern@cs.wisc.edu)
    
   *)


Require Import Max.
Require Import Collapsed_Definitions.

(* ********************************************************************** *)
(** * #<a name="depth"></a># [depth]  *)

Fixpoint depth (t : syntax) :=
  match t with
  | bvar _ => 1
  | fvar _ => 1
  | typ_top => 1
  | typ_arrow t1 t2 => S (max (depth t1) (depth t2))
  | typ_all t1 t2 => S (max (depth t1) (depth t2))
  | typ_sum t1 t2 => S (max (depth t1) (depth t2))  
  | exp_abs t1 t2 => S (max (depth t1) (depth t2))
  | exp_app t1 t2 => S (max (depth t1) (depth t2)) 
  | exp_tabs t1 t2 => S (max (depth t1) (depth t2))
  | exp_tapp t1 t2 => S (max (depth t1) (depth t2))
  | exp_let t1 t2 => S (max (depth t1) (depth t2))
  | exp_inl t1 => S (depth t1)
  | exp_inr t1 => S (depth t1) 
  | exp_case t1 t2 t3 => S (max (depth t1) (max (depth t2) (depth t3))) 
  end
.

(* ********************************************************************** *)
(** * #<a name="properties"></a># Properties of [depth]  *)

(** the depth of a term is always greater than zero. *)

Lemma depth_gt_0 : forall t,
  depth t > 0. 
Proof.
  induction t; simpl; auto with arith.
Qed.


(** opening with a variable preserves depth **)

Lemma depth_open_var : forall k t (z : atom),
  depth t = depth (open_rec k (fvar z) t).
Proof.
  intros k t.
  generalize dependent k.
  induction t; simpl; intros; auto; try solve [f_equal; simpl; auto].
  destruct (k === n); simpl; auto.
Qed.


(** the depth of a term is unchanged if the term is opened with a variable *)

Lemma depth_open_rec : forall n k t (z : atom),
  depth t <= n ->
  depth (open_rec k (fvar z) t) <= n.
Proof.
  intros. rewrite <- depth_open_var. auto.
Qed.

Lemma depth_open : forall n t (z : atom),
  depth t <= n ->
  depth (open t z) <= n.
Proof.
  intros; unfold open; apply depth_open_rec; auto.
Qed.

(* ********************************************************************** *)
(** * #<a name="tactics"></a># Tactics  *)

(** [simpl_ineq] simplifies inequalities about depth that occur in hypotheses *)

Require Import Arith.

Ltac simpl_ineq :=
  match goal with
  | H : S ?t1 <= S ?t2 |- _ => assert (t1 <= t2); [apply le_S_n; apply H | clear H]; simpl_ineq
  | H : max ?t1 ?t2 <= ?t3 |- _ =>
         assert (t1 <= t3 /\ t2 <= t3) as LeConj;
         [split; 
            [eapply le_trans; [eapply le_max_l | eexact H] 
              | eapply le_trans; [eapply le_max_r | eexact H]]
         | clear H; destruct LeConj];
         simpl_ineq
  | _ => idtac
  end.