(** Inductive principles for syntax which match structure of unary predicates

    Author: Anne Mulhern (mulhern@cs.wisc.edu)
    
   *)

Require Import Collapsed_Infrastructure.
Require Import Collapsed_Depth_Binders.


(* ********************************************************************** *)
(** * #<a name="induction principle"></a># induction principle for syntax *)

(** [syntax_rect_cofinite'] is an induction principle on syntax, but follows the
      structure of lc *)

Definition syntax_rect_cofinite' : forall (n : nat) (P : syntax -> Type),
  (forall n : nat, P n) ->
  (forall a : atom, P a) ->
  P typ_top ->
  (forall s : syntax,
    P s -> forall s0 : syntax, P s0 -> P (typ_arrow s s0)) ->
  (forall s : syntax,
    P s -> 
    forall s0 : syntax, (forall x, x `notin` (fv s `union` fv s0) -> P (open s0 x))  -> 
    P (typ_all s s0)) ->
  (forall s : syntax,
    P s -> forall s0 : syntax, P s0 -> P (typ_sum s s0)) ->
  (forall s : syntax,
    P s ->
    forall s0 : syntax, (forall x, x `notin` (fv s `union` fv s0) -> P (open s0 x)) ->
    P (exp_abs s s0)) ->
  (forall s : syntax,
    P s -> forall s0 : syntax, P s0 -> P (exp_app s s0)) ->
  (forall s : syntax,
    P s ->
    forall s0 : syntax, (forall x, x `notin` (fv s `union` fv s0) -> P (open s0 x)) ->
    P (exp_tabs s s0)) ->
  (forall s : syntax,
    P s -> forall s0 : syntax, P s0 -> P (exp_tapp s s0)) ->
  (forall s : syntax,
    P s ->
    forall s0 : syntax, (forall x, x `notin` (fv s `union` fv s0) -> P (open s0 x)) ->
    P (exp_let s s0)) ->
  (forall s : syntax, P s -> P (exp_inl s)) ->
  (forall s : syntax, P s -> P (exp_inr s)) ->
  (forall s : syntax,
    P s ->
    forall s0 : syntax, (forall x, x `notin` (fv s `union` fv s0) -> P (open s0 x)) ->
    forall s1 : syntax, (forall x, x `notin` (fv s `union` fv s0 `union` fv s1) -> P (open s1 x)) ->
    P (exp_case s s0 s1)) ->
  forall (s : syntax), depth s <= n -> P s.
Proof.
  intros. revert H. revert s. revert n.
  induction n.
  Case "0".
    intros s Depth.
    assert (depth s > 0) by auto using depth_gt_0.
    assert False by omega. tauto.
  Case "S n".
    intros s Depth.
    generalize dependent n.
    revert s.
    refine (fix F (s : syntax) := match s as s1 return
      forall n : nat, (forall s0 : syntax, depth s0 <= n -> P s0) -> depth s1 <= S n -> P s1 with
      | bvar n0 => (fun n IHn Depth => X n0)
      | _ => _
      end);
    intros; simpl in *; simpl_ineq; eauto 7 using depth_open.
Qed.

(** [syntax_rect_cofinite] is an induction principle on syntax, but follows the
      structure of lc. It encapsulates facts about the depth of s.  *)

Lemma syntax_rect_cofinite : forall (P : syntax -> Type),
  (forall n : nat, P n) ->
  (forall a : atom, P a) ->
  P typ_top ->
  (forall s : syntax,
    P s -> forall s0 : syntax, P s0 -> P (typ_arrow s s0)) ->
  (forall s : syntax,
    P s -> 
    forall s0 : syntax, (forall x, x `notin` (fv s `union` fv s0) -> P (open s0 x))  -> 
    P (typ_all s s0)) ->
  (forall s : syntax,
    P s -> forall s0 : syntax, P s0 -> P (typ_sum s s0)) ->
  (forall s : syntax,
    P s ->
    forall s0 : syntax, (forall x, x `notin` (fv s `union` fv s0) -> P (open s0 x)) ->
    P (exp_abs s s0)) ->
  (forall s : syntax,
    P s -> forall s0 : syntax, P s0 -> P (exp_app s s0)) ->
  (forall s : syntax,
    P s ->
    forall s0 : syntax, (forall x, x `notin` (fv s `union` fv s0) -> P (open s0 x)) ->
    P (exp_tabs s s0)) ->
  (forall s : syntax,
    P s -> forall s0 : syntax, P s0 -> P (exp_tapp s s0)) ->
  (forall s : syntax,
    P s ->
    forall s0 : syntax, (forall x, x `notin` (fv s `union` fv s0) -> P (open s0 x)) ->
    P (exp_let s s0)) ->
  (forall s : syntax, P s -> P (exp_inl s)) ->
  (forall s : syntax, P s -> P (exp_inr s)) ->
  (forall s : syntax,
    P s ->
    forall s0 : syntax, (forall x, x `notin` (fv s `union` fv s0) -> P (open s0 x)) ->
    forall s1 : syntax, (forall x, x `notin` (fv s `union` fv s0 `union` fv s1) -> P (open s1 x)) ->
    P (exp_case s s0 s1)) ->
  forall (s : syntax),  P s.
Proof.
  intros.
  apply (syntax_rect_cofinite' (depth s)); auto.
Qed.

(** Derived induction principles. *)

Definition syntax_rec_cofinite (P : syntax -> Set) := syntax_rect_cofinite P.

Definition syntax_ind_cofinite (P : syntax -> Prop) := syntax_rect_cofinite P.

(** Renaming lemma for lc *)
Lemma lc_rename : forall x y s,
  x `notin` fv s -> lc (open s x) ->
  y `notin` fv s -> lc (open s y).
Proof.
  intros x y s Frx Wx Fry.
  rewrite (@subst_intro x); auto.
Qed.

(** [lc_dec] can be proved diretcly by induction using syntax_ind_cofinite *)

Lemma lc_dec : forall t,
  lc t \/ ~lc t.
Proof.
  induction t using syntax_ind_cofinite; intros;
    (* break into all recursive cases *)
    repeat
      match goal with
      | H : lc _ \/ ~ lc _ |- _ => destruct H
      | H : forall _, _ `notin` ?L -> lc _ \/ ~lc _ |- _ =>
                let Z := fresh "Z" in
                pick fresh Z; lapply (H Z); auto; intros HZ; destruct HZ; clear H
      end;
    (* solve all cases where term is not locally closed *)
    try solve [
      right; red; intros Inv; inversion Inv; subst;
      match goal with
      | H: ~ lc _ |- False =>
             apply H; let X := fresh "X" in pick fresh X; auto using (lc_rename X)
      | _ => idtac
      end];
     (* solve all cases where term is locally closed *)
      let L := gather_atoms in
      let L := beautify_fset L in
      pose (FreshSet := L);
      left; econstructor; auto;
      try instantiate (1 := FreshSet); intros; subst FreshSet;
      match goal with
        | Z : atom |- _ => let LC := constr:(lc_rename Z) in progress eauto using LC
      end;
    (* solve non-recursive cases *)
     try solve [left; constructor];
     try solve [right; red; intros Inv; inversion Inv].
Qed.