(** Infrastructure and tactics for proof that local closure is decidable by refl

    Author: Anne Mulhern (mulhern@cs.wisc.edu)
    
   *)

Require Import Bool.
Require Import Collapsed_Depth.
Require Import Collapsed_Infrastructure.


(* ********************************************************************** *)
(** * #<a name="local_closure"></a># local closure *)

(** [is_lc_aux n t] decides whether or not a term is locally closed given n, 
      which must be at least the depth of the term *)

Fixpoint is_lc_aux (n : nat) (t : syntax) {struct n} :=
  let new_atom t1 :=
    let (a, _) := atom_fresh_for_set (t1) in
    a in
  match n with
  | 0 => false
  | S n =>
  match t with
  | bvar _ => false
  | fvar _ => true
  | typ_top => true
  | typ_arrow t1 t2 => (is_lc_aux n t1) && (is_lc_aux n t2)
  | typ_all t1 t2 =>
       (is_lc_aux n t1) && (is_lc_aux n (open t2 (new_atom (fv t1 `union` fv t2))))
  | typ_sum t1 t2 => (is_lc_aux n t1) && (is_lc_aux n t2)
  | exp_abs t1 t2 =>
       (is_lc_aux n t1) && (is_lc_aux n (open t2 (new_atom (fv t1 `union` fv t2))))
  | exp_app t1 t2 => (is_lc_aux n t1) && (is_lc_aux n t2)
  | exp_tabs t1 t2 =>
       (is_lc_aux n t1) && (is_lc_aux n (open t2 (new_atom (fv t1 `union` fv t2))))
  | exp_tapp t1 t2 => (is_lc_aux n t1) && (is_lc_aux n t2)
  | exp_let t1 t2 =>
       (is_lc_aux n t1) && (is_lc_aux n (open t2 (new_atom (fv t1 `union` fv t2))))
  | exp_inl t1 => is_lc_aux n t1
  | exp_inr t1 => is_lc_aux n t1
  | exp_case t1 t2 t3 => 
      (is_lc_aux n t1) && 
      (is_lc_aux n (open t2 (new_atom (fv t1 `union` fv t2 `union` fv t3)))) && 
      (is_lc_aux n (open t3 (new_atom (fv t1 `union` fv t2 `union` fv t3))))
  end
  end.

(** [is_lc t] decides whether a term is locally closed *)

Definition is_lc (t : syntax) :=
  let D := depth t in
  is_lc_aux D t.

(** [simpl_bool_conj] simplifies the conjunctions arising in the evaluation if [is_lc_aux] *)
   
Ltac simpl_bool_conj :=
  match goal with
  | H : ?t1 && ?t2 = true |- _ =>
       assert (t1 = true /\ t2 = true) as TConj;
       [apply andb_prop; exact H | clear H; destruct TConj]; simpl_bool_conj
  | H : ?t1 && ?t2 = false |- _ =>
       destruct (andb_false_elim _ _ H); clear H; simpl_bool_conj
  | _ => idtac
  end.

(** [simpl_fresh] simplifies the calculation of a fresh variable in [is_lc_aux] *)
Ltac simpl_fresh :=
  match goal with
  | H : context [atom_fresh_for_set ?L] |- _ =>
           destruct (atom_fresh_for_set L) as [new_atom NewFresh]
  | _ => idtac
  end.


(* ********************************************************************** *)
(** * #<a name="local_closure_dec"></a># local closure is decidable *)

(** if [is_lc_aux n t = true] then [t] is locally closed *)
  
Lemma is_lc_aux_true_lc : forall n t,
  depth t <= n ->
  is_lc_aux n t = true ->
  lc t.
Proof.
  induction n.
  Case "0".
    intros t Depth.
    assert (depth t > 0) by auto using depth_gt_0.
    assert False by omega. tauto.
  Case "S n".
    intros t.
    destruct t; simpl; intros Depth Eq; simpl_bool_conj; simpl_ineq; simpl_fresh;
       try discriminate;
       try solve [constructor; eauto];
       try solve [  let L := gather_atoms in
                         let L := beautify_fset in
                         pose (FreshSet := L);
                         econstructor; 
                         auto;
                         try instantiate (1 := FreshSet);
                         intros X Notin;
                         match goal with
                         | |- lc (open ?t _) => 
                                  assert (lc (open t new_atom)) by auto using depth_open
                         end;
                         rewrite (@subst_intro new_atom); auto].
Qed.

(** if [is_lc t = true] then [t] is locally closed *)

Lemma is_lc_true_lc : forall t,
  is_lc t = true ->
  lc t.
Proof.
  unfold is_lc. intros.
  apply (is_lc_aux_true_lc (depth t)); auto.
Qed.

(** if [is_lc_aux n t = false] then [t] is not locally closed *)

Lemma is_lc_aux_false_not_lc : forall n t,
  depth t <= n ->
  is_lc_aux n t = false ->
  ~lc t.
Proof.
  induction n.
  Case "0".
    intros t Depth.
    assert (depth t > 0) by auto using depth_gt_0.
    assert False by omega. tauto.
  Case "S n".
    intros t.
    destruct t; simpl; intros Depth Eq; red; intros Inv; simpl_ineq;
    try solve [discriminate];   
    try solve [inversion Inv];
    try solve [
      inversion Inv; subst; simpl_fresh; simpl_bool_conj;
      match goal with
      | H : is_lc_aux ?n ?t = false |- _ =>
          lapply (IHn t); auto using depth_open; intros EqNot; apply EqNot; auto
     end;
     pick fresh Z;
     match goal with
     | |- lc (open ?t _) => assert (lc (open t Z))
     end;
     rewrite (@subst_intro Z); auto].    
Qed.

(** if [is_lc t = false] then [t] is not locally closed *)

Lemma is_lc_false_not_lc : forall t,
  is_lc t = false ->
  ~ lc t.
Proof.
  unfold is_lc. intros. 
  apply (is_lc_aux_false_not_lc (depth t)); auto.
Qed.

(** [is_lc t] is either true or false **)

Lemma is_lc_true_false : forall t,
  is_lc t = true \/ is_lc t = false.
Proof.
  intros. destruct (is_lc t); auto.
Qed.

(** local closure is decidable *)

Lemma lc_dec : forall t,
  lc t \/ ~lc t.
Proof.
  intros t.
  destruct (is_lc_true_false t).
  left. apply is_lc_true_lc. auto.
  right. apply is_lc_false_not_lc. auto.
Qed.