(** Infrastructure and tactics for proofs by reflection on a tagged syntax.

    Author: Anne Mulhern (mulhern@cs.wisc.edu)
    
   *)

Require Import Max.
Require Import Tagged_Definitions.


(* ********************************************************************** *)
(** * #<a name="depth"></a># [depth]  *)

Fixpoint depth (T : tag) (t : syntax T) {struct t} :=
  match t with
  | bvar _ _ => 1
  | fvar _ _ => 1
  | typ_top => 1
  | typ_arrow t1 t2 => max (depth _ t1) (depth _ t2)
  | typ_all t1 t2 => S (max (depth _ t1) (depth _ t2))
  | typ_sum t1 t2 => max (depth _ t1) (depth _ t2)  
  | exp_abs t1 t2 => S (max (depth _ t1) (depth _ t2))
  | exp_app t1 t2 => max (depth _ t1) (depth _ t2)
  | exp_tabs t1 t2 => S (max (depth _ t1) (depth _ t2))
  | exp_tapp t1 t2 => max (depth _ t1) (depth _ t2)
  | exp_let t1 t2 => S (max (depth _ t1) (depth _ t2))
  | exp_inl t1 => depth _ t1
  | exp_inr t1 => depth _ t1 
  | exp_case t1 t2 t3 => S (max (depth _ t1) (max (depth _ t2) (depth _ t3))) 
  end
.


(* ********************************************************************** *)
(** * #<a name="properties"></a># Properties of [depth]  *)

(** the depth of a term is always greater than zero. *)

Lemma depth_gt_0 : forall T t,
  depth T t > 0. 
Proof.
  induction t; simpl; eauto with arith.
Qed.

(** opening with a variable preserves depth *)

Lemma depth_open_var : forall k T U t (z : atom),
  depth T t = depth T (open_rec k (@fvar U z) t).
Proof.
  intros k T U t z.
  generalize dependent k.
  induction t; simpl; intros; auto; try solve [f_equal; simpl; auto].
  destruct (k === n); subst; auto.
  destruct U; destruct T; simpl; auto.
Qed.

(** the depth of a term is unchanged if the term is opened with a variable *)

Lemma depth_open_rec : forall n k T U t (z : atom),
  depth T t <= n ->
  depth T (open_rec k (@fvar U z) t) <= n.
Proof.
  intros; rewrite <- depth_open_var; auto.
Qed.

Lemma depth_open : forall n T U t (z : atom),
  depth T t <= n ->
  depth T (open t (@fvar U z)) <= n.
Proof.
  intros; unfold open; apply depth_open_rec; auto.
Qed.

(* ********************************************************************** *)
(** * #<a name="tactics"></a># Tactics  *)

(** [simpl_ineq] simplifies inequalities about depth *)

Require Import Arith.

Ltac simpl_ineq :=
  match goal with
  | H : S ?t1 <= S ?t2 |- _ => assert (t1 <= t2); [apply le_S_n; apply H | clear H]; simpl_ineq
  | H : max ?t1 ?t2 <= ?t3 |- _ =>
         assert (t1 <= t3 /\ t2 <= t3) as LeConj;
         [split; 
            [eapply le_trans; [eapply le_max_l | eexact H] 
              | eapply le_trans; [eapply le_max_r | eexact H]]
         | clear H; destruct LeConj];
         simpl_ineq
  | |- S ?t1 <= S ?t2 => apply le_n_S; simpl_ineq
  | _ => idtac
  end.