(** Inductive principles for syntax which match structure of unary predicates

    Author: Anne Mulhern (mulhern@cs.wisc.edu)
    
   *)

Require Import Tagged_Infrastructure.
Require Import Tagged_Depth.

(* ********************************************************************** *)
(** * #<a name="induction principle"></a># induction principle for syntax *)

(** [syntax_rect_cofinite'] is an induction principle on syntax, but follows the
      structure of lc. It incorporates facts about the depth of s. *)


Definition syntax_rect_cofinite' : forall (n : nat) (P : forall t : tag, syntax t -> Type),
  (forall (T : tag) (n : nat), P T (bvar n)) ->
  (forall (T : tag) (a : atom), P T (fvar a)) ->
  P Typ typ_top ->
  (forall s : syntax Typ,
    P Typ s -> forall s0 : syntax Typ, P Typ s0 -> P Typ (typ_arrow s s0)) ->
  (forall s : syntax Typ,
    P Typ s -> 
    forall s0 : syntax Typ, 
      (forall x, x `notin` (fv s `union` fv s0) -> P Typ (open s0 (@fvar Typ x)))  -> 
    P Typ (typ_all s s0)) ->
  (forall s : syntax Typ,
    P Typ s -> forall s0 : syntax Typ, P Typ s0 -> P Typ (typ_sum s s0)) ->
  (forall s : syntax Typ,
    P Typ s ->
    forall s0 : syntax Exp,
      (forall x, x `notin` (fv s `union` fv s0) -> P Exp (open s0 (@fvar Exp x))) ->
    P Exp (exp_abs s s0)) ->
  (forall s : syntax Exp,
    P Exp s -> forall s0 : syntax Exp, P Exp s0 -> P Exp (exp_app s s0)) ->
  (forall s : syntax Typ,
    P Typ s ->
    forall s0 : syntax Exp,
      (forall x, x `notin` (fv s `union` fv s0) -> P Exp (open s0 (@fvar Typ x))) ->
    P Exp (exp_tabs s s0)) ->
  (forall s : syntax Exp,
    P Exp s -> forall s0 : syntax Typ, P Typ s0 -> P Exp (exp_tapp s s0)) ->
  (forall s : syntax Exp,
    P Exp s ->
    forall s0 : syntax Exp,
      (forall x, x `notin` (fv s `union` fv s0) -> P Exp (open s0 (@fvar Exp x))) ->
    P Exp (exp_let s s0)) ->
  (forall s : syntax Exp, P Exp s -> P Exp (exp_inl s)) ->
  (forall s : syntax Exp, P Exp s -> P Exp (exp_inr s)) ->
  (forall s : syntax Exp,
    P Exp s ->
    forall s0 : syntax Exp,
      (forall x, x `notin` (fv s `union` fv s0) -> P Exp (open s0 (@fvar Exp x))) ->
    forall s1 : syntax Exp,
      (forall x, x `notin` (fv s `union` fv s0 `union` fv s1) -> P Exp (open s1 (@fvar Exp x))) ->
    P Exp (exp_case s s0 s1)) ->
  forall (t : tag) (s : syntax t), depth t s <= n -> P t s.
Proof.
  intros. revert H. revert s. revert t. revert n.
  induction n.
  Case "0".
    intros t s Depth.
    assert (depth t s > 0) by auto using depth_gt_0.
    assert False by omega. tauto.
  Case "S n".
    intros t.
    refine (fun s => match s as s0 in syntax t0 return depth t0 s0 <= S n -> P t0 s0 with
    | bvar t n => fun _ => X t n
    | _ => _
    end);
    intros; simpl in *; simpl_ineq; auto using depth_open.
Qed.

(** [syntax_rect_cofinite] is an induction principle on syntax, but follows the
      structure of lc. It encapsulates facts about the depth of s. *)

Definition syntax_rect_cofinite : forall (P : forall t : tag, syntax t -> Type),
  (forall (T : tag) (n : nat), P T (bvar n)) ->
  (forall (T : tag) (a : atom), P T (fvar a)) ->
  P Typ typ_top ->
  (forall s : syntax Typ,
    P Typ s -> forall s0 : syntax Typ, P Typ s0 -> P Typ (typ_arrow s s0)) ->
  (forall s : syntax Typ,
    P Typ s -> 
    forall s0 : syntax Typ, 
      (forall x, x `notin` (fv s `union` fv s0) -> P Typ (open s0 (@fvar Typ x)))  -> 
    P Typ (typ_all s s0)) ->
  (forall s : syntax Typ,
    P Typ s -> forall s0 : syntax Typ, P Typ s0 -> P Typ (typ_sum s s0)) ->
  (forall s : syntax Typ,
    P Typ s ->
    forall s0 : syntax Exp,
      (forall x, x `notin` (fv s `union` fv s0) -> P Exp (open s0 (@fvar Exp x))) ->
    P Exp (exp_abs s s0)) ->
  (forall s : syntax Exp,
    P Exp s -> forall s0 : syntax Exp, P Exp s0 -> P Exp (exp_app s s0)) ->
  (forall s : syntax Typ,
    P Typ s ->
    forall s0 : syntax Exp,
      (forall x, x `notin` (fv s `union` fv s0) -> P Exp (open s0 (@fvar Typ x))) ->
    P Exp (exp_tabs s s0)) ->
  (forall s : syntax Exp,
    P Exp s -> forall s0 : syntax Typ, P Typ s0 -> P Exp (exp_tapp s s0)) ->
  (forall s : syntax Exp,
    P Exp s ->
    forall s0 : syntax Exp,
      (forall x, x `notin` (fv s `union` fv s0) -> P Exp (open s0 (@fvar Exp x))) ->
    P Exp (exp_let s s0)) ->
  (forall s : syntax Exp, P Exp s -> P Exp (exp_inl s)) ->
  (forall s : syntax Exp, P Exp s -> P Exp (exp_inr s)) ->
  (forall s : syntax Exp,
    P Exp s ->
    forall s0 : syntax Exp,
      (forall x, x `notin` (fv s `union` fv s0) -> P Exp (open s0 (@fvar Exp x))) ->
    forall s1 : syntax Exp,
      (forall x, x `notin` (fv s `union` fv s0 `union` fv s1) -> P Exp (open s1 (@fvar Exp x))) ->
    P Exp (exp_case s s0 s1)) ->
  forall (t : tag) (s : syntax t), P t s.
Proof.
  intros.
  apply (syntax_rect_cofinite' (depth t s)); auto.
Qed.

(** Derived induction principles. *)

Definition syntax_rec_cofinite (P : forall t, syntax t -> Set) := syntax_rect_cofinite P.

Definition syntax_ind_cofinite (P : forall t, syntax t -> Prop) := syntax_rect_cofinite P.

(** Renaming lemma for lc *)
Lemma lc_rename : forall (U T : tag) x y (s : syntax T),
  x `notin` fv s -> lc (open s (@fvar U x)) ->
  y `notin` fv s -> lc (open s (@fvar U y)).
Proof.
  intros U T x y s Frx Wx Fry.
  rewrite (@subst_intro T U x); auto.
Qed.

(** [lc_dec] can be proved directly by induction using syntax_ind_cofinite *)

Lemma lc_dec : forall T t,
  (@lc T t) \/ ~(@lc T t).
Proof with try solve [right; red; intro Inv; inversion Inv; subst; auto].
   intros T s.
   induction s using syntax_ind_cofinite; intros;
   (* break apart into cases *)
    repeat
      match goal with
      | H : lc _ \/ ~ lc _ |- _ => destruct H
      | H : forall _, _ `notin` ?L -> lc _ \/ ~lc _ |- _ =>
                let Z := fresh "Z" in
                pick fresh Z; lapply (H Z); auto; intros HZ; destruct HZ; clear H
      end;
    (* solve cases where t is not locally closed *)
    try solve [
      right; red; intros Inv; inversion Inv; subst;
      match goal with
      | H: ~ lc _ |- False =>
             apply H; let X := fresh "X" in pick fresh X; 
             solve [auto using (lc_rename Typ Typ X) |
                       auto using (lc_rename Typ Exp X) |
                       auto using (lc_rename  Exp Typ X) |
                       auto using (lc_rename Exp Exp X) ]
      | _ => idtac
      end]; 
      (* solve cases where t is locally closed *)
      let L := gather_atoms in
      let L := beautify_fset L in
      pose (FreshSet := L);
      left; econstructor; auto;
      try instantiate (1 := FreshSet); intros; subst FreshSet;
      match goal with
        | Z : atom |- _ => idtac Z;
                solve[
                   auto using (lc_rename Typ Typ Z) |
                   auto using (lc_rename Typ Exp Z) |
                   auto using (lc_rename Exp Typ Z) |
                   auto using (lc_rename Exp Exp Z) ]
       | _ => idtac
    end.
Qed.