(** Infrastructure and tactics for proofs that local closure is decidable via reflection

    Author: Anne Mulhern (mulhern@cs.wisc.edu)
    
   *)

Require Import Max.
Require Import Omega.
Require Import Bool.
Require Import Tagged_Depth.
Require Import Tagged_Infrastructure.


(* ********************************************************************** *)
(** * #<a name="local_closure"></a># local closure *)

(** [is_lc_aux n T t] decides whether or not a term is locally closed given n, 
      which must be at least the depth of the term *)

Fixpoint is_lc_aux (n : nat) (T : tag) (t : syntax T) {struct n} :=
  let new_var t1 T1:=
    let (a, _) := atom_fresh_for_set (t1) in
    (@fvar T1 a) in
  match n with
  | 0 => false
  | S n =>
  match t with
  | bvar _ _  => false
  | fvar _ _ => true
  | typ_top => true
  | typ_arrow t1 t2 => 
      (is_lc_aux n _ t1) && (is_lc_aux n _ t2)
  | typ_all t1 t2 => 
      (is_lc_aux n _ t1) && (is_lc_aux n _ (open t2 (new_var (fv t1 `union` fv t2) Typ)))
  | typ_sum t1 t2 => 
      (is_lc_aux n _ t1) && (is_lc_aux n _ t2)
  | exp_abs t1 t2 => 
      (is_lc_aux n _ t1) && (is_lc_aux n _ (open t2 (new_var (fv t1 `union` fv t2) Exp)))
  | exp_app t1 t2 =>
      (is_lc_aux n _ t1) && (is_lc_aux n _ t2)
  | exp_tabs t1 t2 =>
      (is_lc_aux n _ t1) && (is_lc_aux n _ (open t2 (new_var (fv t1 `union` fv t2) Typ)))
  | exp_tapp t1 t2 =>
      (is_lc_aux n _ t1) && (is_lc_aux n _ t2)
  | exp_let t1 t2 =>
      (is_lc_aux n _ t1) && (is_lc_aux n _ (open t2 (new_var (fv t1 `union` fv t2) Exp)))
  | exp_inl t1 => is_lc_aux n _ t1
  | exp_inr t1 => is_lc_aux n _ t1
  | exp_case t1 t2 t3 => 
      (is_lc_aux n _ t1) && 
      (is_lc_aux n _ (open t2 (new_var (fv t1 `union` fv t2 `union` fv t3) Exp))) && 
      (is_lc_aux n _ (open t3 (new_var (fv t1 `union` fv t2 `union` fv t3) Exp)))
  end
  end.

(** [is_lc T t] decides whether a term is locally closed *)

Definition is_lc (T : tag) (t : syntax T) :=
  let D := depth T t in
  is_lc_aux D T t.

(** [simpl_bool_conj] simplifies the conjunctions arising in the evaluation if [is_lc_aux] *)
   
Ltac simpl_bool_conj :=
  match goal with
  | H : ?t1 && ?t2 = true |- _ =>
       assert (t1 = true /\ t2 = true) as TConj;
       [apply andb_prop; exact H | clear H; destruct TConj]; simpl_bool_conj
  | H : ?t1 && ?t2 = false |- _ =>
       destruct (andb_false_elim _ _ H); clear H; simpl_bool_conj
  | _ => idtac
  end.

(** [simpl_fresh] simplifies the calculation of a fresh variable in [is_lc_aux] *)
Ltac simpl_fresh :=
  match goal with
  | H : context [atom_fresh_for_set ?L] |- _ =>
           pose (FreshSet := L);
           destruct (atom_fresh_for_set L) as [new_atom NewFresh]
  | _ => idtac
  end.


(* ********************************************************************** *)
(** * #<a name="local_closure_dec"></a># local closure is decidable *)

(** if [is_lc_aux n T t = true] then [t] is locally closed *)
  
Lemma is_lc_aux_true_lc : forall n T t,
  depth T t <= n ->
  is_lc_aux n T t = true ->
  lc t.
Proof.
  induction n.
  Case "0".
    intros T t Depth.
    assert (depth T t > 0) by auto using depth_gt_0.
    assert False by omega. tauto.
  Case "S n".
    intros T t.
    destruct t; simpl; intros Depth Eq; simpl_bool_conj; simpl_ineq; simpl_fresh;
       try discriminate;
       try solve [constructor; eauto];
       try solve [  econstructor; 
                         auto;
                         try instantiate (1 := FreshSet);
                         intros X Notin;
                         match goal with
                         | |- lc (open ?t _) => 
                                  try first [
                                    solve [assert (lc (open t (@fvar Exp new_atom)))
                                       by auto using depth_open;
                                     rewrite (@subst_intro Exp Exp new_atom); auto] |
                                    solve [assert (lc (open t (@fvar Typ new_atom)))
                                       by auto using depth_open;
                                     rewrite (@subst_intro Exp Typ new_atom); auto] |  
                                    solve [assert (lc (open t (@fvar Exp new_atom)))
                                       by auto using depth_open;
                                     rewrite (@subst_intro Typ Exp new_atom); auto] |
                                    solve [assert (lc (open t (@fvar Typ new_atom)))
                                       by auto using depth_open;
                                     rewrite (@subst_intro Typ Typ new_atom); auto] ]        
                         end ].
Qed.

(** if [is_lc T t = true] then [t] is locally closed *)

Lemma is_lc_true_lc : forall T t,
  is_lc T t = true ->
  lc t.
Proof.
  unfold is_lc. intros.
  apply (is_lc_aux_true_lc (depth T t)); auto.
Qed.

(** if [is_lc_aux n T t = false] then [t] is not locally closed *)

Lemma is_lc_aux_false_not_lc : forall n T t,
  depth T t <= n ->
  is_lc_aux n T t = false ->
  ~lc t.
Proof.
  induction n.
  Case "0".
    intros T t Depth.
    assert (depth T t > 0) by auto using depth_gt_0.
    assert False by omega. tauto.
  Case "S n".
    intros T t.
    destruct t; simpl; intros Depth Eq; red; intros Inv; simpl_ineq;
    try solve [discriminate];   
    try solve [inversion Inv];
    try solve [
      inversion Inv; subst; simpl_fresh; simpl_bool_conj;
      match goal with
      | H : is_lc_aux ?n ?T ?t = false |- _ =>
          lapply (IHn T t); auto using depth_open; intros EqNot; apply EqNot; auto
      end;
      pick fresh Z;
      match goal with
      | |- lc (open ?t _) => 
           try first [assert (lc (open t (@fvar Typ Z))); auto; 
                           rewrite (@subst_intro Typ Typ Z); auto |
                         assert (lc (open t (@fvar Typ Z))); auto;
                           rewrite (@subst_intro Exp Typ Z); auto |
                         assert (lc (open t (@fvar Exp Z))); auto;
                           rewrite (@subst_intro Typ Exp Z); auto |
                         assert (lc (open t (@fvar Exp Z))); auto;
                           rewrite (@subst_intro Exp Exp Z); auto]
       end].
Qed.

(** if [is_lc T t = false] then [t] is not locally closed *)

Lemma is_lc_false_not_lc : forall T t,
  is_lc T t = false ->
  ~ lc t.
Proof.
  unfold is_lc. intros. 
  apply (is_lc_aux_false_not_lc (depth T t)); auto.
Qed.

(** [is_lc T t] is either true or false **)

Lemma is_lc_true_false : forall T t,
  is_lc T t = true \/ is_lc T t = false.
Proof.
  intros. destruct (is_lc T t); auto.
Qed.

(** local closure is decidable *)

Lemma lc_dec : forall T t,
  @lc T t \/ ~ @lc T t.
Proof.
  intros T t.
  destruct (is_lc_true_false T t).
  left. apply is_lc_true_lc. auto.
  right. apply is_lc_false_not_lc. auto.
Qed.